/* $Header: /usr2/brugere/jbln/uc133/src/RCS/hyper.c,v 1.1 1995/02/17 22:06:47 jbln Exp $ */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <ctype.h>

#include "hyper.h"
#include "uc.h"


/**********************************************************************
   HYPERTEXT VIEWER
   ----------------
   Hypertext format is normal text including the following codes:

      <e,link>		An entry called 'link'
      <l,text,link>	A link to the entry called 'link'
      <h> </h>		A headline (on /off)
      <c> </c>		Center text (on /off)
      <i> </i>		Indent /unindent
      <f> </f>		Preformattet tekst (on /off)
      <z>		A horizontal line
      <b>		A bullet
      <n>		A newline
      <p>		Paragraf end
      <<		The '<'

   Example:

   <e,main>
   This is a Hypertext example.
   This might not come on a newline, because hypertext is a free format
   text. <n>
   This is on a new line.<n>
   So is this. <n>
   This is a lesserthan symbol << !!  <n><n>

   Here comes a bulet
   <b> This is a bullet
   <b> So is this.<n><n>

   < l , click here, link1> to follow a link. <l,and the same here,link2>.<n>
   <b> Links may be placed in <l,bullets,link3><n>

   <e,link1>
   You selected the first link.

   <e,link2>
   You select the second link.

   <e,link3>
   This link could have described bullets ...

**********************************************************************/



typedef struct LINKLIST_T
{
   char link[HT_LINKSIZE+1];
   struct LINKLIST_T *next;
} LINKLIST;

static LINKLIST *ht_stack = NULL;

static char *ht_def_text = {"Entry not found !"};

#define ht_mvaddch(_ht,_y,_x,c) mvaddch(_ht->y+_y, _ht->x+_x, c)

char *ht_show_format();

#define PLMNUMMAX 20

/**************************************************************************/

HTEXT *ht_new(fname, l,c,y,x)
char *fname;
int l,c,y,x;
{
   HTEXT *ht;
   FILE *dummy;

   if (l < 3  ||  c < 10)
      return(NULL);

   if ((dummy=fopen(fname,"r")) == (FILE *)NULL)
      return(NULL);
   else
      fclose(dummy);

   if ((ht=(HTEXT *)malloc(sizeof(HTEXT))) == NULL)
   {
      return(NULL);
   }

   ht->ifile = NULL;
   ht->fname = strdup(fname);
   ht->l = l;
   ht->c = c;
   ht->x = x;
   ht->y = y;

   ht->prevlm = (int *)malloc(sizeof(int) * PLMNUMMAX);

   ht_clear(ht);
   ht_box(ht);

   return(ht);
}


ht_del(ht)
HTEXT *ht;
{
   char dummy[HT_LINKSIZE+1];

   if (ht == NULL)
      return(0);

   fclose(ht->ifile);
   free(ht->fname);
   free(ht);

      /* Clear stack */
   while (ht_pop(dummy))
      /* Nothing */;

   return(1);
}


/**************************************************************************/

ht_run(ht, init_entry)
HTEXT *ht;
char *init_entry;
{
   int c;
   int more=1;
   char entry[HT_LINKSIZE+1];

   if (ht == NULL)
      return(0);

   strcpy(entry, init_entry);

   do
   {
      if (!ht_find(ht, entry))
      {
	 ht->txt = strdup(ht_def_text);
	 ht->linknum = 0;
	 ht->ppnum = 0;
	 ht->cp = ht->ocp = ht->txt;
      }
      else
      {
	 if (!ht_load(ht))
	 {
	    return(0);
	 }
      }

      do
      {
	 if (ht_show(ht) == 0)
            mvprintw(ht->y + ht->l - 1, ht->x + 3, "[more]");
	 else
	    mvprintw(ht->y + ht->l - 1, ht->x + 3, "======");

	 c = ht_edit(ht, entry);

	 if (tolower(c) == tolower(HT_KEYQUIT))
	    more = 0;
      }
      while (c == HT_KEYDOWN  ||  c == HT_KEYUP);

      ht_remove(ht);
   }
   while (more);

   return(1);
}


/**************************************************************************/

static ht_find(ht, entry)
HTEXT *ht;
char *entry;
{
   int c;
   int found=0, more=1;
   char type, name[HT_NAMESIZE+1], link[HT_LINKSIZE+1];

   if ( (ht->ifile=fopen(ht->fname, "r")) == NULL )
      return(0);

   rewind(ht->ifile);

   while (more)
   {
      c = getc(ht->ifile);
      if (c == '<')
      {
	 if (ht_parse(ht, &type, name, link))
	 {
	    if (type == HT_ENTRY)
	    {
	       if (strcmp(link, entry) == 0)
	       {
		  found = 1;
		  more = 0;
	       }
	    }
	 }
      }
      else
      if (c == EOF)
	 more = 0;
   }

   return(found);
}


static ht_load(ht)
HTEXT *ht;
{
   static char fmttxt[5] = {"</f>"};
   int fmtpos=0;
   long fpos;
   int size=0, linknum=0;
   int more=1, n, c;
   char type, name[HT_NAMESIZE+1], link[HT_LINKSIZE+1];
   int preformat;

   fpos = ftell(ht->ifile);

   for (n=0 ; n<2 ; n++)
   {
      size = 0;
      linknum = 0;
      preformat = 0;

      while(more)
      {
	 c = getc(ht->ifile);
	 if (c == '<'  &&  !preformat)
	 {
	    if (ht_parse(ht, &type, name, link))
	    {
	       if (type == HT_LESS)
	       {
		  if (n == 1)
		  {
		     ht->txt[size] = '<';
		  }
		  size += 1;
	       }
	       else
	       if (type == HT_LINK)
	       {
		  if (n == 1)
		  {
		     ht->links[linknum].name = strdup(name);
		     ht->links[linknum].link = strdup(link);
		     ht->txt[size] = HT_EXT;
		     ht->txt[size+1] = HT_LINK;
		     ht->txt[size+2] = linknum;
		  }
		  linknum++;
		  size += 3;
	       }
	       else
	       if (type == HT_ENTRY)
		  more = 0;
	       else
	       {
		  if (n == 1)
		  {
		     ht->txt[size] = HT_EXT;
		     ht->txt[size+1] = type;
		  }
		  size += 2;
		  if (type == HT_FORMAT)
		     preformat = 1;
	       }
	    }
	 }
	 else
	 if (isprint(c)  ||  isspace(c))
	 {
	    if (n == 1)
	       ht->txt[size] = c;
	    size++;
	 }
	 else
	 if (c == EOF)
	    more = 0;

	    /* Try to find </f> to stop preformat */
	 if (preformat  &&  !isspace(c))
	 {
	    if (c == fmttxt[fmtpos])
	       fmtpos++;
	    else
	       fmtpos = 0;
	    if (fmttxt[fmtpos] == 0)
	    {
	       preformat = 0;
	       size -= 4;
	       if (n == 1)
	       {
		  ht->txt[size] = HT_EXT;
		  ht->txt[size+1] = HT_END_FORMAT;
	       }
	       size += 2;
	    }
	 }

      } /* while more */

	 /* Is this first pass ? */
      if (n == 0)
      {
	 fseek(ht->ifile, fpos, SEEK_SET);
	 ht->txt = (char *)malloc(size+1);
	 if (linknum > 0)
	    ht->links = (HTLINK *)malloc(sizeof(HTLINK) * linknum);
	 ht->linknum = linknum;
      }
      more = 1;
      ht->txt[size] = HT_END;

   } /* for n=0 */

   ht->cp = ht->txt;
   ht->ppnum = 0;
   ht->prevpage = NULL;
   ht->preformat = 0;
   ht->lm = 2;
   ht->rm = 2;
   ht->plmnum = 0;

   fclose(ht->ifile);

   return(1);
}


static ht_remove(ht)
HTEXT *ht;
{
   int n;

   for (n=0 ; n<ht->linknum ; n++)
   {
      free(ht->links[n].name);
      free(ht->links[n].link);
   }
   if (ht->linknum > 0)
      free(ht->links);

   free(ht->txt);
   if (ht->ppnum > 0)
      free(ht->prevpage);
}


/**************************************************************************/

static ht_show(ht)
HTEXT *ht;
{
   char *p, *op;
   int more=1, morelink;
   int x,y=1;
   int ln, l, n;
   int blocksize = -1;
   int remark=0;
   int softnl=0;	/* Was last move a soft newline */
   int lastchar=0;

   ht_clear(ht);
   p = ht->cp;
   x = ht->lm;

      /* Clear links */
   for (n=0 ; n<ht->linknum ; n++)
      ht->links[n].visible = 0;

   while (more)
   {
      op = p;

      if (remark)
      {
	 if (*p == HT_EXT)
	 {
	    p++;
	    if (*p == HT_LINK)
	       p++;
	    else
	    if (*p == HT_END_REM)
	       remark = 0;
	    p++;
	 }
	 else
	    p++;
      }
      else
      if (ht->preformat)
      {
	 p = ht_show_format(ht, p, &x, &y, &more);
      }
      else
	 /* Is next symbol an extended one ? */
      if (*p == HT_EXT)
      {
	 p++;

	    /* Is it a remark */
	 if (*p == HT_REM)
	 {
	    p++;
	    remark = 1;
	 }

	 else

	    /**** Is it a bullet or newline ? */
	 if (*p == HT_NEWLINE  ||  *p == HT_BULLET)
	 {
	    if (x!=ht->lm  ||  (*p == HT_NEWLINE && !softnl))
	    {
	       y++;
	       if (y >= ht->l-1)
		  more = 0;
	    }
	    softnl = 0;
	    x = ht->lm;
	    if (*p == HT_BULLET)
	    {
	       if (more)
	       {
		  x = 7;
		  ht_mvaddch(ht, y,4, HT_BULLETSYMB);
		  p++;
	       }
	       else
		  p = op;
	    }
	    else
	       p++;
	 }

	 else

	    /**** Is it a new paragraph ? */
	 if (*p == HT_PARAGRAPH)
	 {
	    p++;
	    x = ht->lm;
	    y += 2;
	    if (y >= ht->l-1)
	       more = 0;
	 }

	 else

	    /**** Is it a horizontal line ? */
	 if (*p == HT_LINE)
	 {
	    p++;
	    softnl = 0;
	    if (x != ht->lm)
	    {
	       y++;
	       if (y >= ht->l-1)
		  more = 0;
	    }
	    if (more)
	    {
	       for (x=2 ; x<ht->c-2 ; x++)
		  ht_mvaddch(ht, y,x, HT_LINESYMB);
	       x = ht->lm;
	       y++;
	       if (y >= ht->l-1)
		  more = 0;
	    }
	 }

	 else

	    /**** Is it a headline */
	 if (*p == HT_HEADLINE)
	 {
	    softnl = 0;
	    p++;
	    if (x != ht->lm)
	    {
	       y++;
	       if (y >= ht->l-1)
		  more = 0;
	    }
	    blocksize = 0;
	 }

	 else

	    /**** Is it a end-headline */
	 if (*p == HT_END_HEADLINE)
	 {
	    softnl = 0;
	    p++;
	    if (x != ht->lm)
	    {
	       y++;
	       if (y >= ht->l-1)
		  more = 0;
	    }
	    if (more)
	    {
	       for (x=2 ; x-2<blocksize  &&  x<ht->c-2 ; x++)
		  ht_mvaddch(ht, y,x, HT_HEADLINESYMB);
	       x=2;
	       y++;
	       if (y >= ht->l-1)
		  more = 0;
	    }
	    blocksize = -1;
	 }

	 else

	    /**** Is it a Center ? */
	 if (*p == HT_CENTER)
	 {
	    p++;
	    l = ht_calclen(p, HT_END_CENTER);
	    if (l>0  &&  l<ht->c-3)
	    {
	       if (ht->c/2 - l/2 > x)
		  x = ht->c/2 - l/2;
	    }
	 }

	 else

	    /**** Is it a End-Center ? */
	 if (*p == HT_END_CENTER)
	 {
	    softnl = 0;
	    p++;
	    x = ht->lm;
	    y++;
	    if (y >= ht->l-1)
	       more = 0;
	 }

	 else

	    /**** Is it a tab ? */
	 if (*p == HT_TAB)
	 {
	    p++;
	    x += HT_TABSIZE - ((x-2) % HT_TABSIZE);
	    if (x >= ht->c - ht->rm)
	       x = ht->rm-1;
	 }

	 else

	    /**** Is it an indent ? */
	 if (*p == HT_INDENT)
	 {
	    p++;
	    l = x + HT_INDENTSIZE - ((x-2) % HT_INDENTSIZE);
	    if (l < ht->l - ht->rm - 1)
	    {
	       ht->prevlm[ht->plmnum++] = ht->lm;
	       ht->lm = l;
	       if (x < ht->lm)
		  x = ht->lm;
	    }
	 }

	 else

	    /**** Is it an end indent ? */
	 if (*p == HT_END_INDENT)
	 {
	    p++;
	       /* Has any indention been done ? */
	    if (ht->plmnum > 0)
	    {
	       ht->plmnum--;
	       ht->lm = ht->prevlm[ht->plmnum];
	       if (x > ht->lm)
		  x = ht->lm;
	    }
	 }

	 else

	    /**** Is it a Pre Format */
	 if (*p == HT_FORMAT)
	 {
	    p++;
	    ht->preformat = 1;
	    p = ht_show_format(ht, p, &x, &y, &more);
	 }

	 else

	    /**** Is it a link */
	 if (*p == HT_LINK)
	 {
	    p++;
	    ln = *p;
	    p++;
	       /* Is this link legal */
	    if (ln>=0  &&  ln<ht->linknum)
	    {
	       ht_bold();
		  /* Is there room on the rest of the line */
	       if (x > ht->c - strlen(ht->links[ln].name) - 2)
	       {
		     /* No - move to next line */
		  softnl = 1;
		  x = ht->lm;
		  y++;
		  if (y >= ht->l - 1)
		     more = 0;
	       }

	       l = 0;
	       morelink = 1;

		  /* If there's room then mark this link as visible */
	       if (more)
	       {
		  ht->links[ln].visible = 1;
		  ht->links[ln].x = x;
		  ht->links[ln].y = y;
	       }

		  /* Write the link name */
	       while (more  &&  morelink)
	       {
		  ht_mvaddch(ht, y,x, ht->links[ln].name[l++]);
		  x++;
		     /* End Of Line ? */
		  if (x >= ht->c - ht->rm)
		  {
		     softnl = 1;
		     x = ht->lm;
		     y++;
		     if (y >= ht->l - 1)
			more = 0;
		  }
		     /* More ? */
		  if (ht->links[ln].name[l] == 0)
		     morelink = 0;
	       }
	       ht_bold_off();
	    }
	 }
	 else	/* Unknown extension */
	    p++;

      }
      else
	 /* Is this the last symbol ? */
      if (*p == HT_END)
      {
	 more = 0;
	 lastchar = 1;
      }
      else
	 /* Is it a leading blank, then remove it */
      if (isspace(*p)  &&  x == ht->lm)
      {
	 p++;
      }
      else
	 /* It is a normal printable symbol */
      {
	    /* Calculate wordlen */
	 l = 0;
	 while (p[l]!=HT_END  &&  p[l]!=HT_EXT  &&  !isspace(p[l]))
	    l++;

	    /* Is there room on the rest of the line ? */
	 if (x > ht->c - l - 2  &&  l>0)
	 {
	    /* No - Insert a newline */
	    x = ht->lm;
	    y++;
	    if (y >= ht->l - 1)
	       more = 0;
	    softnl = 1;
	 }
	 else
	    softnl = 0;

	    /* If there is room for this word then show it */
	 if (more)
	 {
	       /* Display \n \t and others as ' ' */
	    if (isspace(*p))
	    {
	       ht_mvaddch(ht, y,x, ' ');
		  /* Collaps multiple blanks into one */
	       while (isspace(*(p+1))  &&  *(p+1) != HT_EXT)
		  p++;
	    }
	    else
	       ht_mvaddch(ht, y,x, *p);
	    x++;
	    if (blocksize != -1)
	       blocksize++;

	       /* Check EOL */
	    if (x >= ht->c-2)
	    {
	       x = ht->lm;
	       y++;
	       if (y >= ht->l-1)
		  more = 0;
	       softnl = 1;
	    }
	    p++;
	 }
      }
   }

   ht->ocp = ht->cp;
   ht->cp = p;

   return(lastchar);
}


char *ht_show_format(ht, p, x, y, more)
HTEXT *ht;
char *p;
int *x, *y;
int *more;
{
   while (*more  &&  ht->preformat)
   {
      if (*p == HT_EXT)
      {
	 p++;
	 if (*p == HT_END_FORMAT)
	    ht->preformat = 0;
	 p++;
      }
      else
      if (*p == '\n')
      {
	 p++;
	 *x = ht->lm;
	 *y += 1;
	 if (*y >= ht->l-2)
	    *more = 0;
      }
      else
      if (*p == '\r')
      {
	 *x = ht->lm;
      }
      else
      {
	 ht_mvaddch(ht, *y,*x, *p);
	 *x += 1;
	 p++;
	 if (*x >= ht->c-2)
	 {
	    *x = ht->lm;
	    *y += 1;
	    if (*y >= ht->l-2)
	       *more = 0;
	 }
      }
   }

   return(p);
}



/**************************************************************************/

static ht_edit(ht, entry)
HTEXT *ht;
char *entry;
{
   int more=1;
   int c, n;
   int curlink = -1, oldlink;

   if (ht->was_pgup)
   {
	 /* Find last link */
      for (n=ht->linknum-1 ; n>=0 && curlink==-1 ; n--)
         if (ht->links[n].visible)
	    curlink = n;
   }
   else
   {
         /* Find first link */
      for (n=0 ; n<ht->linknum && curlink==-1 ; n++)
         if (ht->links[n].visible)
	    curlink = n;
   }

   ht->was_pgup = 0;

   while (more)
   {
      ht_cursor(ht, curlink, 1);
      refresh();
      c = ht_getch();

         /* Find next link */
      if ((c==HT_KEYNEXT  ||  c==9)  &&  curlink != -1)
      {
	 ht_cursor(ht, curlink, 0);
	 oldlink = curlink;

	 curlink = (curlink+1) % ht->linknum;
	 if (!ht->links[curlink].visible  &&  *(ht->cp) != HT_END)
	    c = HT_KEYDOWN;
	 else
	 {
	    do
	    {
	       if (ht->links[curlink].visible)
	          oldlink = curlink;
	       else
	          curlink = (curlink+1) % ht->linknum;
	    }
	    while (oldlink != curlink);
         }
      }

	 /* Find previously link */
      if (c == HT_KEYPREV  &&  curlink != -1)
      {
	 ht_cursor(ht, curlink, 0);
	 oldlink = curlink;

         curlink--;
         if (curlink < 0)
            curlink = ht->linknum - 1;

	 if (!ht->links[curlink].visible  &&  ht->ppnum>0)
	    c = HT_KEYUP;
	 else
	 {
	    do
	    {
	       if (ht->links[curlink].visible)
	          oldlink = curlink;
	       else
	       {
	          curlink--;
	          if (curlink < 0)
	             curlink = ht->linknum - 1;
	       }
	    }
	    while (oldlink != curlink);
	 }
      }

      if (tolower(c) == tolower(HT_KEYQUIT))
	 more = 0;

      if (c == HT_KEYBACK)
      {
	 if (ht_pop(entry))
	    more = 0;
      }

      if ((c==HT_KEYDOWN || c==' ')  &&  *(ht->cp) != HT_END)
      {
	    /* Push page position */
	 if (ht->ppnum == 0)
	    ht->prevpage = (PAGEINFO *)malloc(sizeof(PAGEINFO));
	 else
	    ht->prevpage = (PAGEINFO *)realloc(ht->prevpage,sizeof(PAGEINFO)*(ht->ppnum+1));
	 ht->prevpage[ht->ppnum].p = ht->ocp;
	 ht->prevpage[ht->ppnum].preformat = ht->preformat;
	 ht->prevpage[ht->ppnum].lm = ht->lm;
	 ht->prevpage[ht->ppnum].rm = ht->rm;
	 ht->ppnum++;

	 more = 0;
      }

      if (c == HT_KEYUP  &&  ht->ppnum > 0)
      {
         ht->was_pgup = 1;
	 ht->ppnum--;
	 ht->cp = ht->prevpage[ht->ppnum].p;
	 if (ht->ppnum > 0)
	 {
	    ht->preformat = ht->prevpage[ht->ppnum-1].preformat;
	    ht->lm = ht->prevpage[ht->ppnum-1].lm;
	    ht->rm = ht->prevpage[ht->ppnum-1].rm;
	 }
	 else
	 {
	    ht->preformat = 0;
	    ht->lm = 2;
	    ht->rm = 2;
	 }
	 if (ht->ppnum == 0)
	    free(ht->prevpage);
	 else
	    ht->prevpage = (PAGEINFO *)realloc(ht->prevpage, sizeof(PAGEINFO) * ht->ppnum);

	 more = 0;
      }

      if ((c==HT_KEYFOLLOW1 || c==HT_KEYFOLLOW2)  &&  curlink != -1)
      {
	 more = 0;
	 ht_push(entry);
	 strcpy(entry, ht->links[curlink].link);
      }
   }

   return(c);
}


static ht_cursor(ht, num, show)
HTEXT *ht;
int num, show;
{
   int l=0,x,y;
   int more=1;

   if (num < 0  ||  num > ht->linknum-1)
   {
      move(ht->y, ht->x);
      return(0);
   }

   if (show)
      ht_reverse();
   else
      ht_bold();
   x = ht->links[num].x;
   y = ht->links[num].y;

   while (ht->links[num].name[l] != 0  &&  more)
   {
      ht_mvaddch(ht, y, x, ht->links[num].name[l]);
      l++;
      x++;
      if (x >= ht->c - 2)
      {
	 x = ht->lm;
	 y++;
	 if (y >= ht->l - 1)
	    more = 0;
      }
   }

   if (show)
      ht_reverse_off();
   else
      ht_bold_off();

   return(1);
}

/**************************************************************************/

static ht_parse(ht, type, name, link)
HTEXT *ht;
char *type, *name, *link;
{
   int l,c;
   int more;
   int endblock=0;

      /* Get type */
   ht_rmblank(ht);
   *type = tolower(getc(ht->ifile));

      /* Test for '/' */
   if (*type == HT_ENDBLOCK)
   {
      endblock = 1;
      *type = tolower(getc(ht->ifile));
   }

   if (*type != HT_LINK  &&  *type != HT_ENTRY)
   {
      name[0] = 0;
      link[0] = 0;

      if (*type == HT_LESS)
	 return(1);

	 /* Remove trailing blanks */
      while ((c=getc(ht->ifile)) != '>'  &&  c != EOF)
	 /* nothing */;

      if (c == EOF)
	 return(0);

      if (endblock)
      {
	 if (*type == HT_CENTER)
	    *type = HT_END_CENTER;
	 if (*type == HT_HEADLINE)
	    *type = HT_END_HEADLINE;
	 if (*type == HT_FORMAT)
	    *type = HT_END_FORMAT;
	 if (*type == HT_INDENT)
	    *type = HT_END_INDENT;
	 if (*type == HT_REM)
	    *type = HT_END_REM;
      }

      return(1);
   }

      /* Get ',' */
   ht_rmblank(ht);
   if (getc(ht->ifile) != ',')
      return(0);

      /* Get name */
   ht_rmblank(ht);
   ht_readstr(ht, name, HT_NAMESIZE);

      /* If it's not a single value entry */
   if (*type != HT_ENTRY)
   {
	 /* Get ',' */
      ht_rmblank(ht);
      if (getc(ht->ifile) != ',')
	 return(0);

	 /* Get link */
      ht_rmblank(ht);
      ht_readstr(ht, link, HT_NAMESIZE);
   }
   else
   {
      strcpy(link, name);
   }

   ht_strlwr(link);

      /* Get '>' */
   ht_rmblank(ht);
   if (getc(ht->ifile) != '>')
      return(0);

   return(1);
}


static ht_rmblank(ht)
HTEXT *ht;
{
   int c;

   do
      c = getc(ht->ifile);
   while (isspace(c));

   ungetc(c, ht->ifile);

   return(1);
}


static ht_readstr(ht, str, size)
HTEXT *ht;
char *str;
int size;
{
   int l=0, more=1, lastspace=0;
   int c;

   while (more  &&  l<size)
   {
      c = getc(ht->ifile);

      if ((isprint(c)  ||  isspace(c))  &&  c!='>'  &&  c!=',')
      {
	 if (!isspace(c))
	    lastspace = 0;
	 else
	 if (lastspace == 0)
	    lastspace = l;

	 str[l++] = c;
      }
      else
      if (c=='>'  ||  c==',')
	 more = 0;
   }

      /* Remove trailing spaces */
   if (lastspace == 0)
      str[l] = 0;
   else
      str[lastspace] = 0;

   ungetc(c, ht->ifile);
   return(1);
}


ht_calclen(p, code)
char *p;
char code;
{
   int found=0;
   int len=0;

   while (*p!=HT_END  &&  !found)
   {
      if (*p == HT_EXT)
      {
	 p++;
	 if (*p == code)
	    found = 1;
	 p++;
      }
      else
      {
	 p++;
	 len++;
      }
   }

   return(len);
}


ht_strlwr(str)
char *str;
{
   while (*str != 0)
   {
      *str = tolower(*str);
      str++;
   }
}

/**************************************************************************/

static ht_push(link)
char *link;
{
   LINKLIST *new;

   new = (LINKLIST *)malloc(sizeof(LINKLIST));
   new->next = ht_stack;
   ht_stack = new;

   strcpy(new->link, link);

   return(1);
}


static ht_pop(link)
char *link;
{
   LINKLIST *next;

   if (ht_stack == NULL)
      return(0);

   strcpy(link, ht_stack->link);

   next = ht_stack->next;
   free(ht_stack);
   ht_stack = next;

   return(1);
}


/**************************************************************************/


static ht_clear(ht)
HTEXT *ht;
{
   int x,y;

   for (x=1 ; x<ht->c-1 ; x++)
      for (y=1 ; y<ht->l-1 ; y++)
	 mvaddch(y+ht->y, x+ht->x, ' ');


   return(1);
}


static ht_box(ht)
HTEXT *ht;
{
   int x,y;

   for (x=0 ; x<ht->c ; x++)
   {
      mvaddch(ht->y, ht->x+x, '=');
      mvaddch(ht->y+ht->l-1, ht->x+x, '=');
   }
   for (y=0 ; y<ht->l ; y++)
   {
      mvaddch(ht->y+y, ht->x, 'I');
      mvaddch(ht->y+y, ht->x+ht->c-1, 'I');
   }
   mvaddch(ht->y, ht->x, '+');
   mvaddch(ht->y+ht->l-1, ht->x, '+');
   mvaddch(ht->y, ht->x+ht->c-1, '+');
   mvaddch(ht->y+ht->l-1, ht->x+ht->c-1, '+');

   return(1);
}


static ht_bold()
{
   attron(A_BOLD);
}


static ht_bold_off()
{
   attroff(A_BOLD);
}


static ht_reverse()
{
   attron(A_REVERSE);
}


static ht_reverse_off()
{
   attroff(A_REVERSE);
}


static int ht_getch()
{
   int c;

   c = getch();
   if (c == 27)
   {
      c = getch();
      if (c == '0')
	 c = HT_KEYQUIT;
   }

   return(c);
}
